#!/usr/bin/env python
# -*- coding: utf-8 -*-
# -*- vim:set ts=4:set noai:syn on:set nu

import re				# Regular Expressions standard module
from types import *		# Some data types
from errors import *	# My error exception data classes
from typecheck import check_types


EOL = "\n"	# End Of Line character

# Private data
E_Unexpected = UnexpectedParameterError # Abbreviation
RESERVED_TOKENS = ['<ERROR>', '<EOF>', '<EPSILON>']	# Reserved tokens. Do not use these
YY_ERROR	= RESERVED_TOKENS[0]
YY_EOF		= RESERVED_TOKENS[1]
YY_EPSILON	= RESERVED_TOKENS[2]

RESERVED_STATES 	= ['YY_INITIAL', 'YY_ERROR', 'YY_NONE', '*']	# Reserved internal status. Do not use these
YY_STATUS_INITIAL 	= RESERVED_STATES[0]
YY_STATUS_ERROR		= RESERVED_STATES[1]
YY_STATUS_NONE		= RESERVED_STATES[2]
YY_STATUS_ANY		= RESERVED_STATES[3] # Universal status


class Token(object):
	''' This class is just used as a C struct to store token values.
	token.text   contains the text recognized from the source code. e.g. '72.500'
	token.value	contains the numerical value of the token when applicable. e.g. 72.5
	token.id		contains the token value for the parser. e.g. 'FLOAT' 
	token.linenum	contains the token linenumber (where it appeared) 
	token.hook		if not None, contains a function that will be called when this 
					token is matched. If the function returns None, then the token
					is discarded, and another token candidate is tried
	'''
	def __init__(self, id, text = '', yy_lineno = None, yy_hook = None, lookahead = None):
		self.id = id
		self.text = text
		self.lineno = yy_lineno
		self.hook = yy_hook
		self.lookahead = lookahead

		try:
			self.value = float(text)
		except ValueError:	# This exceptions occurs if it's not a number
			self.value = None

	def skip(self, lexer):
		''' Tells the lexer to skip this token by returning False.
		'''
		return False

	def reject(self, lexer):
		''' Tells the lexer to reject this match, and try another one.
		'''
		return None

	def __str__(self):
		''' String representation for this token. 
		'''
		return self.id


	def __deepcopy__(self, memo):
		''' Implements this class own deepcopy.
		'''
		result = Token(self.id, self.text, self.lineno, self.hook, self.lookahead)
		result.value = self.value

		return result



class Blex(object):
	''' Class which implements a Lexical Analyzer.
		See the end of the source code for an example of use
	'''

	class Pattern(object):
		''' This class stores token data and patterns definitions
		defined by the user. It's just a container.
		'''
		def __init__(self, pattern, tokenval, caseins = False):
			''' Ensures tokenval is either a string or a tuple(String, function)
			'''
			check_types(StringTypes, pattern, E_Unexpected, 'string')
			check_types((StringTypes, {TupleType:[StringTypes, FunctionType]}), tokenval, E_Unexpected, 'string')

			self.pattern = pattern
			self.re = re.compile(pattern, re.IGNORECASE) if caseins else re.compile(pattern)

			if isinstance(tokenval, TupleType): # Is it a Tuple
				self.hook = tokenval[1]
				tokenval = tokenval[0]
			else:
				self.hook = None

			self.token = tokenval
			

	class Status(object):
		''' Lex status. An status is just a STRING. When the analyzer is on a given status,
		it will ignore what has been defined in the others.
		'''
		def __init__(self, name):
			self.name = name
			self.patterns = {} 
			self.pattern_table = [] # Ordered table of patterns


		def contains_pattern(self, p):
			return p in self.pattern_table

		def add_pattern(self, p):
			''' Adds a pattern object to the pattern list. If already defined,
			en Exception is thrown. '''
			if self.contains_pattern(p): # Already defined?
				raise DuplicatedPatternError(p.pattern)

			self.pattern_table = [p.pattern] + self.pattern_table	# Prepends it
			self.patterns[p.pattern] = p



		
	def __init__(self):
		''' Class constructor. User must call add_token (or add_tokens) before starting
		'''
		self.lineno = 1		# Line number
		self.status_table = {}
		self.add_status(RESERVED_STATES)	# List of available status
		self.set_status(YY_STATUS_INITIAL)	# Current status
		self.set_buffer('')
		self.GREEDY = False
		self.at_newline = True
		self.__lookahead = 0

	
	def set_lookahead(self, lh):
		''' Setter for the lookahead.
		'''
		self.__lookahead = lh
		if lh == 0:
			self.at_newline = True
		else:
			try:
				self.at_newline = self.buffer[lh - 1] == EOL
			except:
				self.at_newline = True

		return lh

		
	def get_lookahead(self):
		return self.__lookahead


	lookahead = property(get_lookahead, set_lookahead)
		

	def set_buffer(self, buffer):
		''' Sets the buffer to the new string.
		'''
		self.lookahead = 0	# Resets the input buffer "pointer"
		self.buffer = buffer
		self.buffer_len = len(buffer)


	def set_status(self, newstatus):
		''' Change current lexer status to the new status. An exception will be 
		thrown if such status is not defined in the status table.
		'''	
		if newstatus not in self.status_table.keys():
			raise UnknownStatusError(newstatus)

		self.status = newstatus


	def add_status(self, status):
		''' Adds a status or a list of them to the list checking its not
		already added.
		'''

		if status == '' or status is None: # "Universal" status
			status = [YY_STATUS_ANY]

		if isinstance(status, StringTypes):
			status = [status]

		check_types('StringListType', status, E_Unexpected, 'string or list of strings')

		for s in status:
			if s not in self.status_table.keys():
				self.status_table[s] = Blex.Status(s)

		return status


	def add_token(self, pattern, tokenval, status = '', caseins = False):
		''' Adds a token to the token table. 
				-pattern: a Regular Expression to recognize the token.
				-tokenval: a string that will be returned when this token is recognized. 
				-status: Blexer STATUS to match this token
				-caseins: True if this pattern is case-insensitive. Defaults to False
		'''
		if self.status_table[YY_STATUS_ANY].contains_pattern(pattern):
			raise DuplicatedPatternError(pattern)
			
		status = self.add_status(status)
		pat = self.Pattern(pattern, tokenval, caseins)
		for s in status:
			self.status_table[s].add_pattern(pat)


	def add_tokens(self, token_map, status = ''):
		check_types(DictType, token_map, E_Unexpected, 'dictionary')

		for token in token_map.keys():
			self.add_token(token, token_map[token], status)
		
	
	def skip_to_eol(self):
		''' Skips input content until End Of Line. End Of Line char is also skipped
		'''
		try:
			self.lookahead += self.buffer[self.lookahead:].index(EOL) + len(EOL)
			self.lineno += 1
			self.at_newline = True
		except ValueError: # Does not exists, and skips until end of file. This was the last line
			self.lookahead = self.buffer_len


	def lex(self):
		''' Funcion que realiza el procesado léxico.
		El buffer de entrada es una cadena almacenada en yy_buffer. '''

		result = None
		while result is None:
			if self.lookahead >= self.buffer_len: # Si ya hemos llegado al final
				return Token(YY_EOF, yy_lineno = self.lineno) # Devolvemos el token 'EOF'
	
			candidate = Token(YY_ERROR, self.buffer[self.lookahead], yy_lineno = self.lineno, lookahead = self.lookahead)
			candidates = [candidate]	# Candidates stack
	
			for status in [self.status, YY_STATUS_ANY]:
				for p in self.status_table[status].pattern_table:
					pattern = self.status_table[status].patterns[p]
					tmp = pattern.re.match(self.buffer[self.lookahead:])
					if tmp is not None:		# Matches?
						candidate = Token(pattern.token, self.buffer[self.lookahead:self.lookahead + tmp.end()], yy_lineno = self.lineno, 
										yy_hook = pattern.hook, lookahead = self.lookahead)
						if not self.GREEDY:	# Not in Greedy mode? then just add it
							candidates = candidates + [candidate]
						else:
							lc = len(candidates)
							for i in range(lc + 1):
								if i == lc or tmp.end() < len(candidates[i].text):
									candidates = candidates[:i] + [candidate] + candidates[i:]
									break
					
			# Saves current lineno and lookahead
			current_lookahead = self.lookahead
			current_lineno = self.lineno
	
			while len(candidates) > 0 and result is None:
				self.lookahead = current_lookahead			# Restore lookahead
				self.lineno	= current_lineno				# Restore lineno
				candidate = candidates.pop()				# The candidate is popped out of the stack
		
				self.lookahead += len(candidate.text) 		# Increases buffer "pointer" the number of chars read
				self.lineno += candidate.text.count(EOL)	# Increases the line counter
				result = candidate
			
				if result.hook is not None:					# has hook?
					result = result.hook(result, self)

			if result == False:
				result = None
			else:
				self.at_newline = result.text[-1] == EOL

		return result


	def look_ahead(self, level = 1):
		''' Return next (lookahead) Token
		'''
		# Saves current lineno and lookahead
		current_lookahead = self.lookahead
		current_lineno = self.lineno
		current_status = self.status

		result = self.lex()
		if (level > 1):
			result = self.look_ahead(level - 1)

		self.lookahead = current_lookahead	# Restore lookahead
		self.lineno = current_lineno		# Restore lineno
		self.status = current_status		# Restore status
		
		return result


