#!/usr/bin/env python
# -*- coding: utf-8 -*-

from types import *
from errors import *
import blexer
from blexer import Blex 
from blexer import Token
from blexer import YY_EOF, YY_ERROR, YY_EPSILON
from typecheck import check_types
from stack import Stack
from ast import Ast

import os
import copy
import re

# Class alias (abbreviation)
E_Unexpected = UnexpectedParameterError

# Operator associativity constants
YY_ASSOC_LEFT = 0
YY_ASSOC_RIGHT = 1


class Storage(object):
	''' Containers Class used for user storage.
	'''
	pass


class Symbol(object):
	''' Symbol class contains symbol name, token, current value, etc.
	It can even used for tree annotation. 
	'''
	def __init__(self, name, value = None, token = None, stype = None, is_terminal = True, lookahead = None):
		self.name = name	# Symbol class. E.g. 'OPERATOR', 'SENTENCE', 'VALUE', etc...
		self.value = value
		self.token = token
		self.type = stype
		self.is_terminal = is_terminal # Whether this symbol is a terminal one or not
		self.firsts = set()  	# FIRST set for this symbol
		self.follows = set()	# Terminal symbols that follows this one
		self.sym_follows = set() # NON-Terminal symbols that follows whis one
		self.lookahead = lookahead
		self.operator = None	# Operator which reduced to this Symbol (should be an Operator instance)
		self.reduced_by = None	# Rule id of which reduced to this Symbol (None for terminal tokens)
		self.epsilons = []		# Rules which has epsilon actions which produced this symbol
		self.ast = None			# Points to the ast node which contains this

		if token is not None:
			if value is None:
				self.value = token.value
	
			if lookahead is None:
				self.lookahead = token.lookahead


class Operator(object):
	''' Class containing operator information.
	'''
	def __init__(self, symbol_id, assoc = YY_ASSOC_LEFT):
		''' Defines operator data.
		'''
		self.id = symbol_id	# ej. '-', '+', '*'
		self.assoc = assoc	# associativity (default: LEFT)
		self.higher = []	# Operators with higher preference than this one
		self.lower = []		# Operators with lower preference than this one


	def takes_lower(self, operator):
		''' Return True if the given operator takes lower precedence
		(less priority) than the given one.
		'''
		return operator.id in self.lower


	def takes_higher(self, operator):
		''' Return True if the given operator takes higher precedence
		(greater priority) than the given one.
		'''
		return operator.id in self.higher



class Rule(object):
	''' Just a container containing a rule definition.
		"hook" is just a function (or None) that will be called whenever the rule is matched.
	'''
	def __init__(self, symbol_name, symbol_list, hook = None, parser = None, epsilon = None):
		check_types('StringType', symbol_name, E_Unexpected, 'string')
		check_types('StringListType', symbol_list, E_Unexpected, 'list of strings')
		
		if len(symbol_list) < 1:
			raise SyntaxError('Nonsense rule: "%s -->"' % symbol_name)

		self.hook = hook # hook ident
		self.left = symbol_name
		self._right = symbol_list
		self.parser = parser
		self.eps_action = epsilon # Action to be done at the beginning of the rule (optional)


	@property
	def right(self):
		return list(self._right)


	def __str__(self):
		result = self.left + ' -->' 
		for i in self.right:
			result += ' ' + i

		return result


	def __call__(self, right_symbols):
		''' Calls the hook with the reduced symbols
		'''
		if self.eps_action is not None:
			del right_symbols[0]

		return self.parser.hooks[self.hook](Symbol(self.left, lookahead = right_symbols[0].lookahead), right_symbols)




class _MicroParser(object):
	''' Private class for our little meta-grammar parser.
	'''
	def __init__(self):
		self.reductors 	= ['-->', '::=']	# Reg. Exp. for a reduction operator
		self.terminator	= ';'				# Chars. defining a Rule terminator
		self.separators = ' \t\n'			# Chars. defining for a separator
		self.counter = 0;					# Counter for in-the-middle actions
		self.pattern = ':SYMBOL%i'			# Symbol pattern
		self.dotre = re.compile(r'(?: +|^)\.([a-zA-Z_][a-zA-Z_0-9]*)?(?: +|$)')	# Regular Exp. to split dot operands
		self.forbidden = re.compile(r'[\[\]/()\\&=<>\-+.,"{}^:;#$@]')		# Forbidden chars in a symbolname
		self.python_ident = re.compile('[a-zA-Z_][a-zA-Z_0-9]*')


	def next_symbol(self):
		result = self.pattern % self.counter
		self.counter += 1

		return result


	def check_valid_symbols(self, symbol):
		''' Raises an exception if symbol chars are not allowed.
		'symbol' can be a string or a list of strings to be checked.
		'''
		if not isinstance(symbol, ListType):
			symbol = [symbol]

		for i in symbol:
			if self.forbidden.search(i):
				raise SyntaxError('Invalid char in symbol name "%s"' % i)

		
	def parse(self, rule):
		''' Parses a rule like:
				for --> FOR IDENT EQ NUMBER TO NUMBER .action
			and returns a tuple containing:
				('for', [ { 'rule'['FOR', 'IDENT', ....] ], 'action')

			a function named action() must be defined before calling bparse.start()

			Expansion #1: Now it allows "." inlined actions, defined as
				for --> FOR IDENT .hook1 EQ NUMBER .hook2 TO NUMBER .hook3 

				This will be translated as:

				for --> TEMP0 TEMP1 TO NUMBER;  (action = hook1)
				TEMP0 --> FOR IDENT; (action = hook3)
				TEMP1 --> EQ NUMBER; (action = hook2)

				Functions hook1(), hook2() and hook3() must be defined before calling bparse.start()
				
				The (optional) number after the dot operator specifies the hook number within the hook list.
				if "." is used with no number, then default_action will be executed.

		'''
		check_types(StringTypes, rule, E_Unexpected, 'string')
		rule = rule.strip(self.separators) 		# Remove whitespaces, tabs and newlines
		rule = rule.rstrip(self.terminator)		# Remove terminating semicolons if any (;)

		left = right = None
		for sep in self.reductors:
			if rule.count(sep):
				left, right = rule.split(sep, 1)
				break
		
		if left is None or right is None:
			raise SyntaxError('Syntax error in grammar. Reduction operator not found, must be one of %s.' % self.reductors)

		left = left.strip(self.separators)	# Clean up separators
		right = right.strip(self.separators)	# Clean up separators

		if left == '':
			raise SyntaxError('Syntax error in grammar. Left rule part is missing')

		if right == '':
			raise SyntaxError('Syntax error in grammar. Right rule part rule is missing')

		rules = right.split('|')	# Disjunction
		result = []
		middle_action_rules = []

		for subrule in rules:
			subrule = subrule.strip(self.separators) # Clean up
			right = []

			_subrules = self.dotre.split(subrule) # Split middle-actions
			subrules = []
			hook_index = []

			if _subrules[-1] == '':
				_subrules.pop()

			if len(_subrules) % 2: 	# Must be even, or add None (default_hook) otherwise
				_subrules.append(None)

			for i in range(0, len(_subrules), 2):
				subrules.append(_subrules[i])
				hook_index.append(_subrules[i + 1])

			for subsubrule, ihook in zip(subrules[:-1], hook_index[:-1]): # From 0 to n - 1
				subsubrule = subsubrule.strip(self.separators).split() # Clean up
				self.check_valid_symbols(subsubrule)

				if subsubrule == []:
					subsubrule = [YY_EPSILON] # Epsilon can only exist at the beginning of the rule
					newsym = YY_EPSILON		  # Create a "fake" rule EPSILON --> EPSILON to be processed later
				else:
					newsym = self.next_symbol() # Allocates a new symbol

				middle_action_rules += [(newsym, subsubrule, ihook)]
				right += [newsym] 

			# Append last part
			subsubrule = subrules[-1].strip(self.separators).split() # Clean up
			self.check_valid_symbols(subsubrule)
			right += subsubrule

			if YY_EPSILON in right: # Has an EPSILON action? (recover the "fake" rule)
				for i in range(len(middle_action_rules)):
					if YY_EPSILON in middle_action_rules[i]:
						epsilon_action = middle_action_rules[i][2]
						middle_action_rules = middle_action_rules[:i] + middle_action_rules[i + 1:] # Remove this fake rule
					break
			else:
				epsilon_action = None

			if [left] == [x for x in right if x != YY_EPSILON] or len(right) < 1:
				raise SyntaxError('Nonsense rule: %s' % Rule(left, right))

			result += [(left, right, hook_index[-1])]
			
		return result + middle_action_rules, epsilon_action
		



class Bparser(Blex):
	''' Implements the parser extending the lexer...
	'''

	class State(object):
		''' Stores the current stack state.
		'''
		def __init__(self, yy_lookahead, yy_stackTOP, stack_len, matches, storage, ast):
			''' Constructor which saves the las N objects of the stack '''
			self.lookahead = yy_lookahead	# Stores current input buffer position
			self.stack_top = copy.deepcopy(yy_stackTOP)
			self.stack_len = stack_len		# len(Bparser.stack) after state push
			self.matches = copy.deepcopy(matches)
			self.storage = copy.deepcopy(storage)
			self.ast = copy.deepcopy(ast)


	class MatchState(object):
		''' Stores which rule has been reduced, so be replayed later (in the mimic phase).

		Parameters:
				rule_id: the rule number to be reduced.
				lookahead: integer which points to the 1st char position of the buffer input
							this rule reduces from the stack.
				eps_action: Set to True to state it was an epsilon pre-reduction
				shift: Set to tokens that where shifted before next reduction
		'''
		def __init__(self, rule_id = None, lookahead = 0, eps_action = False):
			self.id = rule_id
			self.lookahead = lookahead
			self.eps_action = eps_action
			self.shift = () 
		


	def __init__(self):
		''' Bparser CONSTRUCTOR.
		'''
		Blex.__init__(self)
		self.infname = ''	# Filename to parse. '' = stdin
		self.outfname = ''	# Filename to output. '' = stdout
		self.symbols = {}	# Dictionary symbols -> SymbolData
		self.miniparser = _MicroParser()
		self.hooks = {}		# Dictionary of hook functions to call when reduction takes place 'action' -> action()
		self.emmit_allowed = False	# True and instructions will be output.
		self.verbose = False		# False and no msg will be output to stdout
		self.fout = None
		self.rules = []		# Rules
		self.operators = {}	# Operators dictionary. [op] --> { data }. Data is an Operator instance
		self.storage = Storage()
		self.oblivion = True # Set to False if you dont want storage to be cleared on mimic

		''' Grammar rules are stored on three arrays and also in a rules[] array:
			Lefts: SymbolIDs (SymbolNames)
			Rights: Rules lefts parts to be reduced.
			Hooks: Function to call after reduction (symbol.data, rule is passed)

			self.rules[] array contains an array of Rule instances
			
			Left rule part is just a list of symbols. E.g. ['TERM' 'OPERATOR' 'TERM']
			A symbol can have more than one rule.
			If a rule list is [] then it's equivalent to EPSILON (umplemented, and probably never done)
			
		'''
	def default_hook(self, symbol, rights):
		''' Default hook for functions with no hook defined.
		'''
		self.report('Action for rule %s --> %s' % (symbol.name, ' '.join(x.name for x in rights)))
		symbol.value = rights[0].value # $$ = $1
	
		return symbol


	def report(self, msg):
		if self.verbose:
			print "%i: %s" % (self.lineno, msg)


	def warning(self, msg):
		self.report("Warning: %s" % msg)
		

	def set_inputfile(self, fname):
		''' Changes the file to parse, and initializes the buffer.
		'''
		f = open(fname, 'rt', 1)
		buffer = f.read()
		f.close()
		self.infname = fname
		self.set_buffer(buffer)


	def set_outputfile(self, fname, overwrite = False):
		''' Changes the output filename. If force_overwrite == True
			The file will be deleted.
		'''
		self.outfname = fname
		
		if os.path.exists(fname) and not overwrite:
			raise FileAlreadyExistError(fname)

		self.fout = open(outfname, 'wt')


	def add_symbol(self, symbol):
		''' Adds a symbol type to our symbol table.
			This table maps (hash) SYMBOLNAME -> SYMBOLDATA
		'''
		if symbol.name not in self.symbols.keys(): # Not added yet?
			self.report('Added symbol "%s".' % symbol.name)
			self.symbols[symbol.name] = symbol
		

	def add_rule(self, symbol, rule_list, hook = None, epsilon = None):
		''' Adds the reduction rule to the table.
		'''
		check_types(StringTypes, symbol, E_Unexpected, 'string')
		check_types('StringListType', rule_list, E_Unexpected, 'list of strings')
		hook is None or	check_types(StringTypes, hook, E_Unexpected, 'string')
			
		self.add_symbol(Symbol(symbol))
		self.symbols[symbol].is_terminal = False # left-rule symbol is always a nonterminal
		self.report('Symbol "%s" is a nonterminal' % symbol)

		for rule in self.rules:
			if rule_list == rule.right:
				if symbol == rule.left:
					raise DuplicatedRuleError(rule)
				
				self.warning('Reduce/Reduce conflict: Symbol "%s" and "%s" have the same left part.' \
					% (symbol, rule.left))

		self.hooks[hook] = None
		self.rules += [Rule(symbol, rule_list, hook, self, epsilon)]

		# Add the right-part symbols
		for i in rule_list:
			self.add_symbol(Symbol(i))


	def is_viable(self, i):
		''' Returns True if the i-th rule on the stack can be reduced and it's left part
		can go AFTER the top of the stack.
		'''
		r = self.rules[i]
		l = len(r.right) + 1

		if self.yy_stack.can_reduce(r, i):
			return l > len(self.yy_stack) or r.left in self.symbols[self.yy_stack.symbols[-l]].sym_follows
		else:
			if r.eps_action is not None and self.yy_stack.can_reduce_eps(r):
				l -= 1
				return l > len(self.yy_stack) or r.left in self.symbols[self.yy_stack.symbols[-l]].sym_follows

		return False


	def get_rule_operator(self, rule):
		''' Returns the 1st operator found in the rule right part.
		Parameter can be either a rule instance or an integer (rule index)
		'''
		if isinstance(rule, IntType):
			rule = self.rules[rule]			

		opers = self.operators.keys()
		for op in rule.right:
			if op in opers:
				return self.operators[op]

		return None


	def candidates(self):
		''' Returns a list of which rules can be reduced. The list is a vector of integers.
		'''
		result = []
		for i in range(len(self.rules)):
			if self.yy_stack.can_reduce(self.rules[i], i):
				if self.is_viable(i):
					result += [i] 
			else:
				if self.rules[i].eps_action is not None and self.yy_stack.can_reduce_eps(self.rules[i]):
					r = self.rules[i]
					p = len(self.yy_stack) - len(r.right)
					is_candidate = True

					for x in range(p, -1, -1):
						if self.yy_stack.symbols[x] != YY_EPSILON:
							break # Can be a candidate prepending EPSILON

						if self.yy_stack[x].rule == i: # This Rule is already pending to be reduced with epsilon
							is_candidate = False
							break

					if is_candidate and self.is_viable(i):
						result += [i]
				
		''' Ok, now checks operator precedence (Operator is any symbol defined as an operator
		using %LEFT or %RIGHT directives)
		'''
		excluded = []
		for i in result:
			op = self.get_rule_operator(i)
			if op is None:
				continue

			p = len(self.yy_stack) - len(self.rules[i].right) # P is the index in the stack
			if self.rules[i].eps_action is not None: # of the 1st right-part symbol of this rule
				p += 1

			q = p
			assoc = YY_ASSOC_LEFT
			for p in range(p, len(self.yy_stack)):
				symbol = self.yy_stack[p]
				if symbol.name == op.id: # Now, after the operator must be RIGHT only
					assoc = YY_ASSOC_RIGHT
					continue

				if symbol.operator is None: # Look the operator which reduced this symbol, if any
					continue # it was not an op-reduction, ignore

				if symbol.operator.id in op.lower: 	# Previous reduction must be rejected:
					excluded.append(i)				# this operator takes higher precedence
					break

				# Now check if the operator is right or left associative
				if symbol.operator.id == op.id and op.assoc != assoc: # Should be only left assoc
					excluded.append(i)
					break

	 	result = [x for x in result if x not in excluded]	
		self.report('Possible candidates: %i' % len(result))

		return result	


	def _stack_report(self, str = ''):
		self.report('%sStack now is: %s (%i)' % (str, self.yy_stack, len(self.yy_stack)))


	def push_state(self, N, i):
		''' Pushes the top N states into the metastack, and also memorizes
		which rule i'th was used.
		'''
		self._stack_report()
		self.report('Pushing current state...')

		if len(self.yy_stack) < N:
			raise StackError('Stack underflow. Trying to pop %i elements, only %i.' % (N, len(self.yy_stack)))

		result = copy.deepcopy(self.yy_stack[-N:])
		self.metastack += [Bparser.State(self.lookahead, self.yy_stack[-N:], len(self.yy_stack), self.matches, self.storage, self.ast)]
		self.yy_stack[-N:] = []		# Pops last N from the stack
		self.linenos.append(self.lineno)
		self._stack_report('Done. ')
		self.__storages[-1][1] = self.metastack[-1].storage # Pointer to the storage copy
		self.__storages.append([set([]), self.storage])

		if not self.mimic_phase:	# If not in the mimic phase
			self.matches.append(Bparser.MatchState(rule_id = i, lookahead = None))	# push the matched rule id for the mimic phase

		return result


	def pop_state(self):
		''' Pops the last stack from the metastack
		'''
		self._stack_report()
		self.report('Recovering previous state...')

		state = self.metastack.pop()	# Pop last state
		self.lookahead = state.lookahead

		N = state.stack_len - len(state.stack_top)

		if N < 0:
			raise StackError('Stack incoherence. Old stack_len %i, len(stack_top) %i.' % (state.stack_len, len(state.stack_top)))

		self.yy_stack[N:] = state.stack_top
		self.lineno = self.linenos.pop()
		self.matches = state.matches
		self.storage = state.storage
		self.ast = state.ast
		self._stack_report('Done. ')
		self.__storages.pop()


	def parse_operator(self, rule_str):
		''' Check if rule_str is an operator definition.
		Operator can be a Token or even a non-terminal symbol.
		You can define associativity and precedence using %LEFT and %RIGHT.
		Latest operators takes higher precedence.
		By default operators are left-associative and have the same priority.
		
		To define a left associative operator with the same priority (ej. '+' and '-') use:
		%LEFT (PLUS, MINUS)  

		Here you declare PLUS and MINUS with the same priority and left associative. Now
		declare a higher precedence operator, MUL and DIV:
		%LEFT (MUL, DIV)

		Finally declare a RIGHT associative operator, POWER (POW), with higher precedence:
		%RIGHT (POW)
		'''
		rule_str = rule_str.strip()
		
		if re.match(r'%(LEFT|RIGHT)[\t\n ]+\([^,]+(,[ \t\n]*[^,]+)*\)$', rule_str) is None: # Try to match a OP definition
			return None # Not mached, must be a rule. Returns None

		opers = [x.strip() for x in rule_str[re.match('[^(]+', rule_str).end() + 1:-1].strip().split(',')] # Gets parameters 
		assoc = YY_ASSOC_LEFT if rule_str[:5] == '%LEFT' else YY_ASSOC_RIGHT

		operators = []
		previous = self.operators.keys()

		for i in opers:
			if i in self.operators.keys():
				raise DuplicatedOperatorError(i)

			self.operators[i] = Operator(i, assoc) # Store it in or operator database
			operators.append(i)

		# Now establish precedence: Operators previously defined takes lower
		for i in previous:
			self.operators[i].higher += operators

		for i in operators:
			self.operators[i].lower += previous
				
		return operators


	def parse_rule(self, rule_str):
		''' Parse a rule like:
			
			for_statement --> FOR IDENT EQUAL NUMBER TO NUMBER ;

			statement --> if_statement | for_statement 
				| other_statement;
						
			For meta-grammar syntax, see examples and docs.
			Basically: 

				--> is the reduction operator (definition) (you can also use ::= instead of --> operator)
				|	is the disjunction operator (you can define more than one rule for a symbol)
				;	ends the rule:
				
			Literal symbols like '+', 'for' etc... are not allowed. They will be treated as symbols/tokens
			It's recomended to use lowercase for non-terminal symbols (like in the previous example)

			You can insert in-the-middle actions using the dot ".". E.g.

			if_statement --> IF .ident1 expr .ident2 THEN statement .ident3;

			Where ident1, ident2 and ident3 are python function identifiers (actions). In the above case,
			two in-the-middle actions ident1(), ident2() and a rule reduction action ident3(). 

			This is equivalent to:

			if_statement --> if expr THEN statement .ident3; 
			if --> IF .ident1;
			expr --> expr_definition .ident2;
		'''	
		if self.parse_operator(rule_str) is not None: # Tries first to check it's an operator definition
			return
		
		rule_info, epsilon_action = self.miniparser.parse(rule_str)

		for rule_def in rule_info:
			symbol, right_part, hook = rule_def
			eps = (YY_EPSILON in right_part and epsilon_action) or None
			self.add_rule(symbol, right_part, hook, eps)


	def shift(self, token = None):
		''' SHIFT the input and stores it onto the stack top.
		If a "token" is specified in the parameter, then the input remains
		unchanged and that token is shifted and pushed onto the stack.
		This is useful to push <EPSILON> tokens.
		'''
		if token is None:
			token = self.lex()				# Ask for next token to the lexer

		if token.id == YY_EOF:			# end of input?
			return None					# No more tokens...

		self.yy_stack += [Symbol(token.id, token = token)]
		self.report('Shifting [%s]' % token.id)

		if self.__storages == []: # Marks current lookahead if not done
			self.__storages.append([set([token.lookahead]), self.storage])
		else:
			self.__storages[-1][0].add(token.lookahead)

		return token.id


	def discard_matches(self, lookahead, rules = []):
		''' Discard matches which has lookahead 
		greater to the given one, or equal to it if the rule which uses it
		is NOT in the rules [] list.
		'''
		self.report('Discarding matches with lookahead > %i or with rules not in %s' % (lookahead, rules))

		for i in range(len(self.matches) - 1, -1, -1):
			if self.matches[i].lookahead < lookahead:
				break

			if self.matches[i].lookahead == lookahead and self.matches[i].eps_action and self.matches[i].id not in rules:
				break

			del self.matches[i]


	def recover_storage(self, old_lh):
		''' Go "backs in time", and recover storage when the lookahead position
		was old_lh
		'''
		for state in self.__storages:
			if state is None:
				continue

			if old_lh in state[0]:
				self.storage = copy.deepcopy(state[1])
				return

		raise StorageRecoverError(old_lh)


	def reduce(self, i):
		''' Reduces stack by rule i-th. 
			Pop the right part symbols out of the stack, then
			push the right part.
		'''
		self.report("Reducing by rule %i: %s" % (i, self.rules[i]))

		# Check if Epsilon is missing in the Stack
		if self.rules[i].eps_action is not None and not self.yy_stack.can_reduce(self.rules[i], i):
			if not self.mimic_phase:
				self.report("Trying (%i: %s) with %s" % (i, self.rules[i], YY_EPSILON))
				right_symbols = self.push_state(len(self.rules[i].right) - 1, i) # Saves current state 
				symbol = Symbol(YY_EPSILON, lookahead = right_symbols[0].lookahead)
				symbol.rule = i # Anotates which rules MUST reduce this EPSILON, if any
				symbol.epsilons += [i]
				self.lookahead = symbol.lookahead # Retry the input
				self.recover_storage(self.lookahead)
				self.matches[-1].lookahead = symbol.lookahead
				symbol = self.hooks[self.rules[i].eps_action](symbol, []) # Calls the hook for this EPSILON
				if symbol is not None:
					lower_rules = [x.reduced_by for x in right_symbols if x is not None ]
					self.discard_matches(symbol.lookahead, right_symbols[0].epsilons) # Tag rules with lookahead >= this one
					self.matches.append(Bparser.MatchState(rule_id = i, lookahead = symbol.lookahead, eps_action = True)) 
			else:
				symbol = self.hooks[self.rules[i].eps_action](Symbol(YY_EPSILON), [])
				if symbol is not None:
					symbol.rule = i
					symbol.epsilons += [i]
		else:
			right_symbols = self.push_state(len(self.rules[i].right), i) # Saves current state 
			self.matches[-1].lookahead = right_symbols[0].lookahead
			symbol = self.rules[i](right_symbols) # Call it with the reduced symbols
			if symbol is not None:
				for x in right_symbols:
					symbol.epsilons += x.epsilons

				if self.rules[i].eps_action is not None:
					symbol.epsilons.append(i)

		if symbol is None: 	# If the action returns None, then ...
			return None		# this reduction is invalidated (returning None to the parser)

		symbol.reduced_by = i		# Saves with rule reduced to this symbol
		symbol.operator = self.get_rule_operator(i)
		self.yy_stack += [symbol]   # Push the produced sumbol into the stack
		self._stack_report('Done. ')

		return symbol


	def emmit(self, str):
		''' Output str. This function is called to output compiled instructions.
		User *MUST* overload this function to output instructions, for example, to a file.'''
		if not self.emmit_allowed:
			return

		if self.fout is None:
			print str,
		else:
			fout.write(str + EOL)


	def firsts(self, symbol):
		''' Calculates the FIRSTs of a given symbol, and stores them in the symbol table.
		Other FIRSTs will be calculated too.
		'''
		# The FIRSTS of a terminal is the terminal itself
		if symbol.is_terminal:
			symbol.firsts = set([symbol.name])
			return symbol.firsts

		for rule in self.rules: # For every rule in the grammar...
			if rule.left != symbol.name: continue # This rule does not define this symbol

			s = rule.right[0] # Take the 1st symbol of the right side of the rule
			if s == YY_EPSILON: # If there is an implicit <EPSILON> take the 2nd symbol
				symbol.firsts = symbol.firsts.union(self.firsts(self.symbols[YY_EPSILON]))
				s = rule.right[1]

			if s in symbol.visited: continue # The 1st are already included
			if s == symbol.name: continue	 # The 1st cannot be itself. For epsilon rules => calculate FOLLOWs

			symbol.visited += [s]
			symbol.firsts = symbol.firsts.union(self.firsts(self.symbols[s]))
		
		return symbol.firsts		


	def follows(self, symbol):
		''' Calculates which terminal symbols follows the given one.
		'''
		if symbol.name in symbol.visited:
			return symbol.follows

		symbol.visited.append(symbol.name)

		for rule in self.rules:
			for i in range(len(rule.right) - 1):
				if symbol.name == rule.right[i]:
					symbol.follows = symbol.follows.union(self.symbols[rule.right[i + 1]].firsts)

			if symbol.name == rule.right[-1]: # Rule === A -> B1 B2 B3 ... S  (?)
				symbol.follows = symbol.follows.union(self.follows(self.symbols[rule.left]))
				
		return symbol.follows


	def sym_follows(self, symbol):
		''' Calculates which non-terminal symbols follows the given one.
		'''
		for rule in self.rules:
			for i in range(len(rule.right) - 1):
				if symbol.name == rule.right[i]:
					symbol.sym_follows = symbol.sym_follows.union([rule.right[i + 1]])

		changed = True
		while changed:
			changed = False

			for rule in self.rules:
				if rule.left not in symbol.sym_follows:
					continue

				if rule.right[0] != YY_EPSILON and rule.right[0] in symbol.sym_follows:
					continue

				if rule.right[0] == YY_EPSILON and rule.right[0] in symbol.sym_follows and \
					rule.right[1] in symbol.sym_follows:
					continue

				symbol.sym_follows = symbol.sym_follows.union([rule.right[0]])
				if rule.right[0] == YY_EPSILON:
					symbol.sym_follows = symbol.sym_follows.union([rule.right[1]])

				changed = True
				
		return symbol.sym_follows

	
	def init_state(self):
		''' Initializes several state variables.
		'''
		self.lookahead = 0
		self.lineno = 1
		self.linenos = []
		self.yy_stack = Stack()
		self.__cutlevel = None
		self.__storages = []
		self.ast = None
		self.metastack = []	# A stack of stacks. This will save compiler state for backtracking


	def start(self):
		''' Initializes the parser:
		Fills the FIRSTS set for each symbol.
		'''
		self.init_state()
		self.epsilon_rules = []     # Array of stack states which might introduce EPSILON tokens
		self.abort = False
		self.mimic_phase = False	# True on the 2nd phase.
		self.emmit_allowed = False 
		self.matches = []	# An array containing the matched rules

		for i in self.symbols.keys():
			self.symbols[i].visited = []

		for i in self.symbols.keys():
			self.firsts(self.symbols[i]) # Calculates the 1sts of each symbol
			self.report('Firsts(%s): %s' % (i, self.symbols[i].firsts))

		for i in self.symbols.keys():
			self.symbols[i].visited = []

		for i in self.symbols.keys(): # Calculate FOLLOWS
			self.follows(self.symbols[i])
			self.report('Follow(%s): %s' % (i, self.symbols[i].follows))

		for i in self.symbols.keys(): # Calculate FOLLOWS
			self.sym_follows(self.symbols[i])
			self.report('SYM_Follows(%s): %s' % (i, self.symbols[i].sym_follows))

		for i in range(len(self.rules)):
			if self.rules[i].eps_action is not None:
				self.hooks[self.rules[i].eps_action] = None

			self.report("Rule %i: %s" % (i, self.rules[i]))

		for i in self.operators.keys():
			op = self.operators[i]
			self.report("Operator '%s': Assoc: %i, lower: %s, higher: %s" % (i, op.assoc, op.lower, op.higher))

		for i in self.hooks.keys(): # Convert hooks to functions if not done already
			if i is None:
				self.hooks[i] = self.default_hook
				continue

			try:
				self.hooks[i] = eval('self.%s' % i)
			except AttributeError:
				raise UndefinedFunctionError(i)

			if not callable(self.hooks[i]):
				raise NotAFunctionError(i)
	

	def backtracking_cut(self):
		self.__cutlevel = self.__recurse_level


	@property
	def recurse_level(self):
		return self.__recurse_level


	def syntax_error(self, lineno, token):
		''' Called when a syntax error has happened.
		Line number and token is passed. Other parameters and data should
		be obtained from the self instance. Overload this function to
		manage your errors.
		'''
		self.report("Syntax error at line %i: Unexpected token %s '%s'" % (lineno, token, token.text))


	def __parse(self, level = 0):
		''' Implements the parser.
		Tries FIRST to reduce by any rule. If it does, then push the state, and call itself recursively.
		Otherwise, shift a token from the input and tries again, until all the input has been matched.
		Level = Recursion Level
		Cut	  = Max recursion level
		'''
		if self.abort:
			return False

		if self.__cutlevel is not None and level <= self.__cutlevel:
			self.report("Cut: At level %i <= %i" % (level, self.__cutlevel))
			self.syntax_error(self.look_ahead(2).lineno, self.look_ahead(2))
			self.abort = True
			return False

		self.__recurse_level = level

		la = self.look_ahead().id
		if len(self.yy_stack) > 0 and la != YY_EOF and la not in self.symbols[self.yy_stack.symbols[-1]].follows:	
			return False

		if len(self.yy_stack) > 1 and self.yy_stack.symbols[-1] == YY_EPSILON and \
			YY_EPSILON not in self.symbols[self.yy_stack.symbols[-2]].follows:
			return False

		candidates = self.candidates() # Posible reductions
		for i in candidates:	# Candidates rules
			self.report('Level %i -- Trying %i: %s' % (level, i, self.rules[i]))

			result = self.reduce(i)
			if result is None:
				self.warning("Rule %s rejected by user hook" % self.rules[i])
			elif result != YY_EPSILON and self.__parse(level + 1):
				break
	
			if self.abort: return False
			self.pop_state() # Recovers previous state and try another rule

		while self.shift() is not None:
			if self.yy_stack.symbols[-1] == YY_ERROR:
				self.syntax_error(self.lineno, self.yy_stack[-1].token)
				self.abort = True
				return False

			self.matches.append(Bparser.MatchState(lookahead = self.yy_stack[-1].lookahead))
			self.matches[-1].shift += (self.yy_stack.symbols[-1], )
			if self.candidates() != [] and self.__parse(level + 1):
				break

			if self.abort: return False

		return self.yy_stack.symbols == [self.rules[0].left]	# Syntax ERROR (no alternatives) unless everything is reduced

	
	def parse(self):
		result = self.__parse()
		if self.oblivion:
			self.storage = Storage()

		return result


	def mimic(self, matches = None):
		''' Repeats the parsing process, executing the actions in the hooks.
		'''
		self.init_state()
		self.mimic_phase = True
		self.emmit_allowed = True 
		index = 0

		self.report("Starting mimic phase...")

		if matches is not None: # Overrides matches in the instance with the ones in the parameter
			self.matches = matches

		while index < len(self.matches):
			status = self.matches[index]

			if status.eps_action:
				self.reduce(status.id)
			elif status.id is not None:	
				rule = status.id
				if rule in self.candidates():
					self.report("Mimic rule %s" % self.rules[rule])
					self.reduce(rule)
				else:
					raise StackError("Incoherent match state with reduction %i: %s" % (rule, self.rules[rule]))

			for i in status.shift:
				tmp = self.shift()
				if tmp != i:
					raise StackError("Incoherent match state. Expected %s, got %s." % (i, tmp))

			index += 1
		

