#!/usr/bin/env python
# -*- coding: utf-8 -*-

# --------------------------- ERROR exception classes ----------------------------

class Error(Exception):
	'''Base class for exceptions in this module.
	'''
	def __init__(self, msg = 'Unknown error'):
		self.message = msg

	def __str__(self):
		return self.message


class UnexpectedParameterError(Error):
	'''Exception raised for errors in parameters.

	Attributes:
		unexecpected_value -- Parameter received
	'''
	def __init__(self, unexpected_value, expected_type):
		self.value = unexpected_value
		self.message = "Unexpected '%s' parameter. A %s was expected instead" % (str(unexpected_value), expected_type)


class DuplicatedPatternError(Error):
	'''Exception raised for errors in patterns.

	Attributes:
		Pattern -- String contained the pattern (Reg. Exp.)
	'''
	def __init__(self, pattern):
		self.pattern = pattern
		self.message = 'Pattern "%s" already defined.' % str(pattern)


class DuplicatedRuleError(Error):
	'''Exception raised for errors in rules (duplicated rule).

	Attributes:
		Rule -- String contained the rule (string or Rule instance)
	'''
	def __init__(self, rule):
		self.rule = rule 
		self.message = 'Rule "%s" already defined.' % str(rule)


class DuplicatedOperatorError(Error):
	'''Exception raised for errors in operator (duplicated operator).

	Attributes:
		Oper -- String contained the Operator (Reg. Exp.)
	'''
	def __init__(self, oper):
		self.oper = oper
		self.message = 'Operator "%s" already defined.' % str(oper)


class DuplicatedTokenError(Error):
	'''Exception raised for errors in parameters.

	Attributes:
		token -- Value of the duplicated token (string)
	'''
	def __init__(self, token):
		self.toke = token 
		self.message = 'Token "%s" already defined.' % str(token)


class SyntaxError(Error):
	'''Exception raised for syntax errors in grammar or parsed file.

	Attributes:
		token -- Value of the duplicated token (string)
	'''
	pass


class StackError(Error):
	'''Exception raised for stack errors (inconsistence, stack overflow, etc)

	Attributes:
		token -- Value of the duplicated token (string)
	'''
	pass


class UnknownStatusError(Error):
	'''Exception thrown when an unknown status has been requested.
	'''
	def __init__(self, status):
		self.status = status
		self.message = 'Unknown status. Status "%s" is not defined.' % status


class FileAlreadyExistError(Error):
	'''Exception thrown when an existing file is about to be overwritten.
	'''
	def __init__(self, fname):
		self.fname = fname
		self.message = "File '%s' already exist. Use overwrite = True or delete it." % fname


class NotAFunctionError(Error):
	'''Exception thrown when an given action ident. is not a python funcion.
	'''
	def __init__(self, ident):
		self.ident = ident
		self.message = "'%s' is not a python function." % ident


class UndefinedFunctionError(Error):
	'''Exception thrown when an action ident is not defined yet.
	'''
	def __init__(self, ident):
		self.ident = ident
		self.message = '"%s" is not defined' % ident


class StorageRecoverError(Error):
	'''Exception thrown when storage couldn't been recovered, or
	it's corrupted.
	'''
	def __init__(self, lookahead):
		self.ident = lookahead
		self.message = 'Storage for lookahead(%i) could not be recovered, or does not exists.' % lookahead
